import inspect
import logging
from dataclasses import dataclass
from typing import List


LOG = logging.getLogger(__name__)


# Similar to MeasurementDescription from metloom
@dataclass(eq=True, frozen=True)
class MeasurementDescription:
    """
    data class for describing a measurement
    """

    code: str = "-1"  # code used within the applicable API
    description: str = None  # description of the sensor
    map_from: List = None  # map to this variable from a list of options
    remap: bool = False  # Auto remap to the column to the code
    # TODO: optional unit we get form 'parse_from' method
    # unit: str = None


class ExtendableVariables:
    """
    Make a class with the helpful iterator portion of enums, but
    that is fully extendable
    """
    def __init__(self, entries=None):
        self._entries = None
        # allows filtering to a desired list
        self._initial_entries = entries or self._all_variables

    @property
    def entries(self):
        if self._entries is None:
            self._entries = self._get_members()
        return self._entries

    @property
    def _all_variables(self):
        return [m[1] for m in self.entries]

    @property
    def variables(self):
        return [m[1] for m in self.entries if m[1] in self._initial_entries]

    @property
    def names(self):
        return [m[0] for m in self.entries if m[1] in self._initial_entries]

    @classmethod
    def _get_members(cls):
        """
        https://www.geeksforgeeks.org/how-to-get-a-list-of-class-attributes-in-python/
        Important this is a class method or seg faults
        """
        result = []
        # getmembers() returns all the
        # members of an object
        for i in inspect.getmembers(cls):

            # to remove private and protected
            # functions
            if not i[0].startswith('_'):

                # To remove other methods that
                # doesnot start with a underscore
                if not inspect.ismethod(i[1]) and not isinstance(i[1], property):
                    result.append(i)
        return result

    def __iter__(self):
        for item in self.variables:
            yield item

    def __len__(self):
        return len(self.entries)


class ProfileVariables(ExtendableVariables):
    @classmethod
    def from_mapping(cls, input_name):
        """
        Get the measurement description from an input name.
        This will use the  MeasurementDescription.map_from list to
        check, in order, which variable we should use

        Returns:
            column name
            column mapping (map of name to MeasurementDescription)
        """
        result = None
        # Map column name to variable type
        column_mapping = {}
        for entry in cls():
            if entry.map_from and input_name.lower() in entry.map_from:
                # Remap to code
                if entry.remap:
                    result = entry.code
                else:
                    result = input_name
                column_mapping[result] = entry
                break
        if result is None:
            raise RuntimeError(f"Could not find mapping for {input_name}")
        LOG.debug(
            f"Mapping {result} to {result} (type {column_mapping[result]})"
        )
        return result, column_mapping

    SWE = MeasurementDescription(
        "SWE", "Snow Water Equivalent",
        ["swe_mm", "swe"]
    )
    DEPTH = MeasurementDescription(
        "depth", "top or center depth of measurement",
        [
            "depth", "top", "sample_top_height", "hs",
            "depth_m", 'snowdepthfilter(m)', 'snowdepthfilter'
        ], True
    )
    BOTTOM_DEPTH = MeasurementDescription(
        "bottom_depth", "Lower edge of measurement",
        ["bottom", "bottom_depth"], True
    )
    DENSITY = MeasurementDescription(
        "density", "measured snow density",
        [
            "density", "density_a", "density_b", "density_c", "avg_density",
            "avgdensity", 'density_mean'
        ]
    )
    LAYER_THICKNESS = MeasurementDescription(
        "layer_thickness", "thickness of layer"
    )
    SNOW_TEMPERATURE = MeasurementDescription(
        "snow_temperature", "Snowpack Temperature",
        ["temperature", "temperature_deg_c"]
    )
    LWC = MeasurementDescription(
        "liquid_water_content", "Liquid water content",
        ["lwc_vol_a", "lwc_vol_b", "lwc", "lwc_vol"]
    )
    PERMITTIVITY = MeasurementDescription(
        "permittivity", "Permittivity",
        ["permittivity_a", "permittivity_b", "permittivity", 'dielectric_constant']
    )
    GRAIN_SIZE = MeasurementDescription(
        "grain_size", "Grain Size",
        ["grain_size"]
    )
    GRAIN_TYPE = MeasurementDescription(
        "grain_type", "Grain Type",
        ["grain_type"]
    )
    HAND_HARDNESS = MeasurementDescription(
        "hand_hardness", "Hand Hardness",
        ["hand_hardness"]
    )
    MANUAL_WETNESS = MeasurementDescription(
        "manual_wetness", "Manual Wetness",
        ["manual_wetness"]
    )


class SnowExProfileVariables(ProfileVariables):
    COMMENTS = MeasurementDescription(
        "comments", "Comments",
        ["comments"]
    )
    PIT_COMMENTS = MeasurementDescription(
        "pit_comments", "Pit Comments",
        ["pit_comments"]
    )
    COUNT = MeasurementDescription(
        "count", "Count for surrounding perimeter depths",
        ["count"]
    )
    PARAMETER_CODES = MeasurementDescription(
        "parameter_codes", "Parameter Codes",
        ["parameter_codes"]
    )
    TIME_BOUND_PIT = MeasurementDescription(
        "Time start/end",
        "Time of first or last pit measurement",
        ["Time start/end", "time_start/end"]
    )
    SITE_NAME = MeasurementDescription(
        "site_name", "Name of campaign site",
        ['location'], True
    )
    SITE_ID = MeasurementDescription(
        "site_id", "ID within campaign site",
        ['site'], True
    )
    PIT_ID = MeasurementDescription(
        "pit_id", "ID of snow pit",
        ['pitid'], True
    )
    EQUIVALENT_DIAMETER = MeasurementDescription(
        'equivalent_diameter', "",
        ['deq']
    )
    OBSERVERS = MeasurementDescription(
        'observers', "Observer(s) of the measurement",
        ['operator', 'surveyors', 'observer']
    )
    TOTAL_DEPTH = MeasurementDescription(
        'total_depth', "Total depth of measurement",
        ['total_snow_depth']
    )
    INSTRUMENT = MeasurementDescription(
        'instrument', "Instrument of measurement",
        ['smp_serial_number', 'measurement_tool', 'instrument']
    )
    LATITUDE = MeasurementDescription(
        'latitude', "Latitude",
        ['lat', 'latitude']
    )
    LONGITUDE = MeasurementDescription(
        'longitude', "Longitude",
        ['long', 'lon', 'longitude']
    )
    EASTING = MeasurementDescription(
        'easting', "UTM Easting",
        ['easting']
    )
    NORTHING = MeasurementDescription(
        'northing', "UTM Northing",
        ['northing']
    )
    TWO_WAY_TRAVEL = MeasurementDescription(
        'two_way_travel', "Two way travel",
        ['twt', 'twt_ns']
    )
    UTM_ZONE = MeasurementDescription(
        'utm_zone', "UTM Zone",
        ['utmzone', 'utm_zone']
    )
    FLAGS = MeasurementDescription(
        'flags', "Measurements flags",
        ['flag']
    )
    DATE = MeasurementDescription(
        'date', "Measurement Date (only date column)",
        ['date_dd_mmm_yy']
    )
    TIME = MeasurementDescription(
        'time', "Measurement time",
        ['time_gmt']
    )
    ELEVATION = MeasurementDescription(
        'elevation', "Elevation",
        ['elev_m']
    )
    DATETIME = MeasurementDescription(
        'datetime', "Combined date and time",
        ["Date/Local Standard Time", "date/local_standard_time", "datetime"],
        True
    )
    RH_10FT = MeasurementDescription(
        "relative_humidity_10ft",
        "Relative humidity measured at 10 ft tower level",
        ['rh_10ft']
    )
    BP = MeasurementDescription(
        'barometric_pressure', "Barometric pressure",
        ['bp_kpa_avg']
    )
    AIR_TEMP_10FT = MeasurementDescription(
        'air_temperature_10ft',
        "Air temperature measured at 10 ft tower level",
        ['airtc_10ft_avg']
    )
    WIND_SPEED_10FT = MeasurementDescription(
        'wind_speed_10ft',
        "Vector mean wind speed measured at 10 ft tower level",
        ['wsms_10ft_avg']
    )
    WIND_DIR_10ft = MeasurementDescription(
        'wind_direction_10ft',
        "Vector mean wind direction measured at 10 ft tower level",
        ['winddir_10ft_d1_wvt']
    )
    SW_IN = MeasurementDescription(
        'incoming_shortwave',
        "Shortwave radiation measured with upward-facing sensor",
        ['sup_avg']
    )
    SW_OUT = MeasurementDescription(
        'outgoing_shortwave',
        "Shortwave radiation measured with downward-facing sensor",
        ['sdn_avg']
    )
    LW_IN = MeasurementDescription(
        'incoming_longwave',
        "Longwave radiation measured with upward-facing sensor",
        ['lupco_avg']
    )
    LW_OUT = MeasurementDescription(
        'outgoing_longwave',
        "Longwave radiation measured with downward-facing sensor",
        ['ldnco_avg']
    )
    SM_20CM = MeasurementDescription(
        'soil_moisture_20cm', "Soil moisture measured at 10 cm below the soil",
        ['sm_20cm_avg']
    )
    ST_20CM = MeasurementDescription(
        'soil_temperature_20cm',
        "Soil temperature measured at 10 cm below the soil",
        ['tc_20cm_avg']
    )



import pytest

# from insitupy.campaigns.variables import ProfileVariables


class TestVariables:
    # TODO: these test
    @pytest.mark.skip("Not done yet")
    def test_variable_attributes(self):
        pass

    def test_variable_list(self):
        pass



"""
Point data from select manual measurement campaigns
"""
import logging
from pathlib import Path
import geopandas as gpd
from typing import List

import numpy as np
import pandas as pd

from .metadata import MetaDataParser, ProfileMetaData
from .variables import ProfileVariables, MeasurementDescription, \
    SnowExProfileVariables

"""

# Start with this
https://github.com/SnowEx/snowex_db/blob/main/scripts/upload/add_time_series_pits.py
data from https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_TS_SP.001/2020.04.27/

# TODO: What do we need here

https://github.com/SnowEx/snowex_db/blob/main/scripts/upload/add_iop_pits.py
https://github.com/SnowEx/snowex_db/blob/main/scripts/upload/add_sbb_depths.py
https://github.com/SnowEx/snowex_db/blob/main/scripts/upload/add_snow_depths.py

SMP
    https://github.com/SnowEx/snowex_db/blob/main/scripts/upload/add_smp.py
    https://github.com/SnowEx/snowex_db/blob/main/scripts/upload/resample_smp.py

Maybe not
    https://github.com/SnowEx/snowex_db/blob/main/scripts/upload/add_snow_poles.py


"""

SOURCES = [
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_SSA.001/",
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_BSU_GPR.001/",
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_GM_SP.001/",
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_SMP.001/",
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_SD.001/",
    ("https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_GM_CSU_GPR.001/2020.02.06/"
     "SNEX20_GM_CSU_GPR_1GHz_v01.csv"),
    ("https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_UNM_GPR.001/2020.01.28/"
     "SNEX20_UNM_GPR.csv"),
    ("https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_SD_TLI.001/2019.09.29/"
     "SNEX20_SD_TLI_clean.csv"),
]


"""
Start with pit density, temperature measurements

Big pluses
    * variable name mapping

# TODO
    * will need login info for NSIDC
    * stratigraphy will be the most complicated

# TODO Next
    * Clean up the metadata file reading
        * We don't need everything we're reading right now
    * map to sample_a, sample_b when reading in file
    * flush out density and swe calcs

"""

LOG = logging.getLogger(__name__)


class ProfileData:
    """
    This would be one pit, SMP profile, etc
    Unique date, location, variable
    """
    VARIABLES = ProfileVariables
    META_PARSER = MetaDataParser

    def __init__(
        self, input_df, metadata: ProfileMetaData, variable: MeasurementDescription,
    ):
        """
        Take df of layered data (SMP, pit, etc)
        Args:
            input_df: dataframe of data
                Should include depth and optional bottom depth
                Should include sample or sample_a, sample_b, etc

        """
        self._depth_layer = self.VARIABLES.DEPTH
        self._lower_depth_layer = self.VARIABLES.BOTTOM_DEPTH
        self._metadata = metadata
        # TODO: auto parse variable if not given
        self.variable: MeasurementDescription = variable
        # mapping of column name to measurement type
        self._column_mappings = {}
        # List of measurements to keep
        self._measurements_to_keep = [
            self._depth_layer, self._lower_depth_layer, self.variable
        ]

        self._id = metadata.id
        self._dt = metadata.date_time

        # This will populate the column mapping
        self._df = self._format_df(input_df)

        columns = self._df.columns.values
        if self._depth_layer.code not in columns:
            raise ValueError(f"Expected {self._depth_layer} in columns")

        # List of columns that are not the desired variable
        _non_measure_columns = [
            self._depth_layer.code, self._lower_depth_layer.code,
            "datetime",
            "geometry"
        ]
        self._non_measure_columns = [
            c for c in _non_measure_columns if c in columns
        ]

        # Columns related to the variable
        self._sample_columns = [
            c for c in columns if self._column_mappings.get(c) == self.variable
        ]
        if len(self._sample_columns) == 0:
            raise ValueError(f"No sample columns in {columns}")

        # describe the data a bit
        self._has_layers = self._lower_depth_layer.code in columns
        # More than 1 sample of the variable (sample_1, sample_2)...
        self._multi_sample = len(self._sample_columns) > 1
        # Extend the df info
        self._extend_df()

    def _format_df(self, input_df):
        """
        Format the incoming df with the column headers and other info we want
        """

        # Get rid of columns we don't want and populate column mapping
        columns = input_df.columns.values
        for c in columns:
            cn, cm = self.VARIABLES.from_mapping(c)
            # join with existing mappings
            self._column_mappings = {**cm, **self._column_mappings}

        columns_to_keep = [
            c for c in columns
            if self._column_mappings[c] in self._measurements_to_keep
        ]
        df = input_df.loc[:, columns_to_keep]

        n_entries = len(df)
        df["datetime"] = [self._dt] * n_entries

        # parse the location
        lat, lon = self.latlon
        location = gpd.points_from_xy(
            [lon] * n_entries, [lat] * n_entries
        )

        df = gpd.GeoDataFrame(
            df, geometry=location
        ).set_crs("EPSG:4326")
        df = df.replace(-9999, np.NaN)

        return df

    def _extend_df(self):
        # set the thickness of the layer
        if self._has_layers:
            self._df[self.VARIABLES.LAYER_THICKNESS.code] = (
                self._df[self._depth_layer.code] - self._df[
                    self._lower_depth_layer.code
                ]
            )

    @property
    def latlon(self):
        # return location metadata
        return self._metadata.latitude, self._metadata.longitude

    @property
    def df(self):
        return self._df

    @property
    def sum(self):
        # get bulk value
        if not self._has_layers:
            # could we assume equidistant spacing and do a mean?
            raise RuntimeError("Cannot compute for no layers")

        # this should work for multi or not multi sample
        profile_average = self._df.loc[:, self._sample_columns].mean(axis=1)
        self._df["mean"] = profile_average
        # TODO: sum up with depth change
        # TODO: could we use the weighted mean * the total depth?
        # TODO: units
        pass

    @property
    def mean(self):
        profile_average = self._df.loc[:, self._sample_columns].mean(
            axis=1)
        if pd.isna(profile_average).all():
            return np.nan
        if self._has_layers:
            # height weighted mean for these layers
            thickness = self._df[self.VARIABLES.LAYER_THICKNESS.code]
            # this works for a weighted mean, but is not assumed to be
            # the total thickness of the snowpack
            thickness_total = thickness.sum()
            weighted_mean = (
                profile_average * thickness / thickness_total
            ).sum()
            value = weighted_mean
        else:
            value = np.mean(profile_average)

        return value

    @property
    def total_depth(self):
        profile = self._df.loc[:, self._depth_layer.code].values
        return np.nanmax(profile)

    def get_profile(self, snow_datum="ground"):
        # TODO: snow datum is ground or snow
        # get profile of values
        profile_average = self._df.loc[:, self._sample_columns].mean(
            axis=1)
        df = self._df.copy()
        df[self.variable.code] = profile_average
        columns_of_interest = [*self._non_measure_columns, self.variable.code]
        return df.loc[:, columns_of_interest]

    @classmethod
    def from_file(self, fname, variable: ProfileVariables):
        raise NotImplementedError("Not implemented")


class SnowExMetadataParser(MetaDataParser):
    VARIABLES_CLASS = SnowExProfileVariables


class SnowExProfileData(ProfileData):
    VARIABLES = SnowExProfileVariables
    META_PARSER = SnowExMetadataParser

    @classmethod
    def from_file(cls, fname, variable: MeasurementDescription):
        # TODO: timezone here (mapped from site?)
        meta_parser = cls.META_PARSER(fname, "US/Mountain")
        # Parse the metadata and column info
        metadata, columns, header_pos = meta_parser.parse()
        # read in the actual data
        data = cls._read(fname, columns, header_pos)

        return cls(data, metadata, variable)

    @staticmethod
    def _read(profile_filename, columns, header_position):
        """
        # TODO: better name mapping here
        Read in a profile file. Managing the number of lines to skip and
        adjusting column names

        Args:
            profile_filename: Filename containing the a manually measured
                             profile
        Returns:
            df: pd.dataframe contain csv data with standardized column names
        """
        # header=0 because docs say to if using skip rows and columns
        df = pd.read_csv(
            profile_filename, header=0,
            skiprows=header_position,
            names=columns,
            encoding='latin'
        )
        # Special SMP specific tasks
        depth_fmt = 'snow_height'
        is_smp = False
        if 'force' in df.columns:
            # Convert depth from mm to cm
            df['depth'] = df['depth'].div(10)
            is_smp = True
            # Make the data negative from snow surface
            depth_fmt = 'surface_datum'

            # SMP serial number and original filename for provenance to the comment
            f = Path(profile_filename).name
            serial_no = f.split('SMP_')[-1][1:3]

            df['comments'] = f"fname = {f}, " \
                             f"serial no. = {serial_no}"

        if not df.empty:
            # Standardize all depth data
            new_depth = standardize_depth(
                df['depth'], desired_format=depth_fmt, is_smp=is_smp
            )

            if 'bottom_depth' in df.columns:
                delta = df['depth'] - new_depth
                df['bottom_depth'] = df['bottom_depth'] - delta

            df['depth'] = new_depth

            delta = abs(df['depth'].max() - df['depth'].min())
            LOG.debug(
                f'File contains a profile with'
                f' with {len(df)} layers across {delta:0.2f} cm'
            )
        return df


class ProfileDataCollection:
    """
    This could be a collection of pits, profiles, etc
    """

    def __init__(self, df):
        self._df = df
        pass

    @property
    def SWE(self):
        """
        Takes all layers for each unique date, location and returns point swe
        geodataframe
        We can iterate though all ProfileData in this class and get SWE DF for each
        """
        # find all points with variable == density and calc SWE
        pass

    def get_mean(self, variable: MeasurementDescription):
        raise NotImplementedError()

    def get_sum(self, variable: MeasurementDescription):
        raise NotImplementedError()

    def get_profile(self, variable: MeasurementDescription):
        raise NotImplementedError()

    # def get_pit_data(self, start_date, end_date, variables) -> List[PointData]:
    #     """
    #     Returns a geodataframe
    #     """
    #     pass

    @property
    def metadata(self):
        # might be a map of date to location
        pass

    @classmethod
    def points_from_geometry(
        cls,
        geometry: gpd.GeoDataFrame,
        variables: List[MeasurementDescription],
        snow_courses=False,
        within_geometry=True,
        buffer=0.0
    ):
        pass

    @classmethod
    def from_files(cls):
        # parse mlutiple files and create an iterable of ProfileData
        pass


def standardize_depth(depths, desired_format='snow_height', is_smp=False):
    """
    Data that is a function of depth comes in 2 formats. Sometimes 0 is
    the snow surface, sometimes 0 is the ground. This function standardizes it
    for each profile. desired_format can be:

        snow_height: Zero at the bottom of the data.
        surface_datum: Zero at the top of the data and uses negative depths
                       (easier for plotting)

    Args:
        depths: Pandas series of depths in either format
        desired_format: string indicating which format the data is in
        is_smp: Boolean indicating which data this is, if smp then the data is
                surface_datum but with positive depths
   Returns:
        new:
    """
    max_depth = depths.max()
    min_depth = depths.min()

    new = depths.copy()

    # How is the depth ordered
    # max_depth_at_top = depths.iloc[0] > depths.iloc[-1]

    # Is the data in surface_datum already
    bottom_is_negative = depths.iloc[-1] < 0

    if desired_format == 'snow_height':

        if is_smp:
            LOG.info('Converting SMP depths to snow height format.')
            new = (depths - max_depth).abs()

        elif bottom_is_negative:
            LOG.info('Converting depths in surface datum to snow height format.')

            new = (depths + abs(min_depth))

    elif desired_format == 'surface_datum':
        if is_smp:
            LOG.info('Converting SMP depths to surface datum format.')
            new = depths.mul(-1)

        elif not bottom_is_negative:
            LOG.info('Converting depths in snow height to surface datum format.')
            new = depths - max_depth

    else:
        raise ValueError(
            f'{desired_format} is an invalid depth format! Options are:'
            f' {["snow_height", "surface_datum"]}'
        )

    return new



import numpy as np
import pytest

from insitupy.campaigns.campaign import SnowExProfileData
from insitupy.campaigns.variables import ProfileVariables


class TestSnowexPitProfile:
    """
    Test the attributes of the profile
    """

    @pytest.mark.parametrize(
        "fname, variable, expected", [
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_temperature_v01.csv",
                ProfileVariables.SNOW_TEMPERATURE, 0.0
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                ProfileVariables.LWC, np.nan
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                ProfileVariables.PERMITTIVITY, np.nan
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv",
                ProfileVariables.DENSITY, 395.037037
            ),
        ]
    )
    def test_mean(self, fname, variable, expected, data_path):
        file_path = data_path.joinpath(fname)
        obj = SnowExProfileData.from_file(file_path, variable)
        result = obj.mean
        if np.isnan(expected):
            assert np.isnan(result)
        else:
            assert result == pytest.approx(expected)

    @pytest.mark.parametrize(
        "fname, variable, expected", [
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_temperature_v01.csv",
                ProfileVariables.SNOW_TEMPERATURE, 95.0
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                ProfileVariables.LWC, 95.0
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                ProfileVariables.PERMITTIVITY, 95.0
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv",
                ProfileVariables.DENSITY, 95.0
            ),
        ]
    )
    def test_total_depth(self, fname, variable, expected, data_path):
        file_path = data_path.joinpath(fname)
        obj = SnowExProfileData.from_file(file_path, variable)
        result = obj.total_depth
        assert result == pytest.approx(expected)
